/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "operator/operator.pmh"

using namespace fe;

void ConnectOp::initialize(void)
{
	catalog<String>("icon")="FE_beta";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Driver Surface");
	catalog<bool>("Driver Surface","optional")=true;
}

void ConnectOp::handle(Record& a_rSignal)
{
	sp<SurfaceAccessorI> spOutputPoint;
	if(!accessOutput(spOutputPoint,a_rSignal,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputPrimitive;
	if(!accessOutput(spOutputPrimitive,a_rSignal,
			e_primitive,e_vertices)) return;

	sp<SurfaceI> spDriver;
	sp<SurfaceAccessorI> spInputUV;
	sp<SurfaceAccessorI> spInputNormal;
	if(!access(spDriver,"Driver Surface",e_quiet))
	{
		if(!access(spInputUV,"Input Surface",e_point,"uv")) return;
		access(spInputNormal,"Input Surface",e_point,e_normal);
	}

	//* collect points
	const U32 outputCount=spOutputPoint->count();
	std::vector<Vector2> locationArray(outputCount);
	std::vector<SpatialVector> normalArray(outputCount);
	if(spDriver.isValid())
	{
		for(U32 pointIndex=0;pointIndex<outputCount;pointIndex++)
		{
			const SpatialVector point=spOutputPoint->spatialVector(pointIndex);
			sp<SurfaceI::ImpactI> spImpact=spDriver->nearestPoint(point);
			if(spImpact.isValid())
			{
				locationArray[pointIndex]=spImpact->uv();
				normalArray[pointIndex]=spImpact->normal();
			}
			else
			{
				set(locationArray[pointIndex],point[0],point[2]);
				normalArray[pointIndex]=spInputNormal.isValid()?
						spInputNormal->spatialVector(pointIndex):
						unitSafe(point);
			}
		}
	}
	else
	{
		for(U32 pointIndex=0;pointIndex<outputCount;pointIndex++)
		{
			locationArray[pointIndex]=spInputUV->spatialVector(pointIndex);
			normalArray[pointIndex]=spInputNormal.isValid()?
					spInputNormal->spatialVector(pointIndex):
					unitSafe(spOutputPoint->spatialVector(pointIndex));
		}
	}

	//* Delaunay triangulation
	std::vector<Vector3i> triangleArray;
	PointConnect::solve(locationArray,triangleArray);

	//* append one primitive per triangle
	const U32 primitiveCount=triangleArray.size();
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const Vector3i& rTriangle=triangleArray[primitiveIndex];
		const I32 p0=rTriangle[0];
		const I32 p1=rTriangle[1];
		const I32 p2=rTriangle[2];

		const SpatialVector point0=spOutputPoint->spatialVector(p0);
		const SpatialVector point1=spOutputPoint->spatialVector(p1);
		const SpatialVector point2=spOutputPoint->spatialVector(p2);
		const SpatialVector triNorm=
				unitSafe(cross(point1-point0,point2-point0));
		const SpatialVector aveNorm=unitSafe(normalArray[p0]+
				normalArray[p1]+normalArray[p2]);
		const BWORD flip=(dot(triNorm,aveNorm)<0.0);

		const I32 createdIndex=spOutputPrimitive->append();
		spOutputPrimitive->append(createdIndex,p0);
		spOutputPrimitive->append(createdIndex,flip? p1: p2);
		spOutputPrimitive->append(createdIndex,flip? p2: p1);
	}
}
