/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __operator_FollicleOp_h__
#define __operator_FollicleOp_h__

namespace fe
{

/**************************************************************************//**
	@brief Operator to edit surface locations

	@ingroup operator
*//***************************************************************************/
class FE_DL_EXPORT FollicleOp:
	public OperatorState,
	public Initialize<FollicleOp>
{
	public:

						FollicleOp(void):
							m_brushed(FALSE),
							m_lastX(-1),
							m_lastY(-1),
							m_lastUV(-1,-1),
							m_dragging(FALSE),
							m_areaSelect(FALSE)								{}

virtual					~FollicleOp(void)									{}

		void			initialize(void);

						//* As HandlerI
virtual	void			handle(Record& a_rSignal);

	protected:

		void			addFollicle(const SpatialVector& a_intersection);

		Record			getEdit(I32 a_index);
		void			movePicked(const SpatialVector& a_rDelta);
		void			slideOnUV(I32 a_index,const Vector2 a_deltaUV);
		void			snapToSurface(I32 a_index);

		void			paint(Real a_intensity);
		BWORD			revert(I32 a_index);
		BWORD			hideToggle(I32 a_index);
		BWORD			hidden(I32 a_index);
		void			hide(I32 a_index);

		BWORD			moved(I32 a_index);

virtual	void			setupState(void);
virtual	BWORD			loadState(const String& rBuffer);

	private:

		sp<SurfaceAccessorI>		m_spInputPoint;
		sp<SurfaceAccessorI>		m_spOutputPoint;
		sp<SurfaceAccessorI>		m_spOutputNormal;
		sp<SurfaceAccessorI>		m_spOutputAttribute;
		sp<SurfaceAccessorI>		m_spHideGroup;
		sp<SurfaceI>				m_spDriver;
		WindowEvent					m_event;

		sp<DrawMode>				m_spWireframe;
		sp<DrawMode>				m_spSolid;

		sp< RecordMap<I32> >		m_spEditMap;
		Accessor<SpatialVector>		m_aLocation;
		Accessor<Vector2>			m_aUV;
		Accessor<Real>				m_aValue;

		BWORD						m_brushed;
		I32							m_lastX;
		I32							m_lastY;
		Vector2						m_lastUV;

		std::vector<Color>			m_colorRim;
		std::vector<Color>			m_colorCenter;
		BWORD						m_dragging;
		BWORD						m_areaSelect;
};

} /* namespace */

#endif /* __operator_FollicleOp_h__ */
